package scripty;

import scripty.Script.RunBuilder;
import scripty.Script.RunBuilder.ArgBuilder;

import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import java.util.*;
import org.xml.sax.SAXException;

public class Parser {

    private final Document document;
    private final Map<String, Script> scripts = new HashMap<>();

    public Parser(InputStream in) throws IOException, ParserConfigurationException, SAXException {
	this.document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
    }

    public Parser(Document document) {
	this.document = document;
    }

    public Program parse() throws ParseException {
	String main = assertAttribute(document.getDocumentElement(), "main");
	Program program = new Program(main);
	List<Element> children = getElementChildren(document.getDocumentElement());
	for (Element child : children) {
	    String tag = child.getTagName();
	    if (tag.equals("bash")) {
		parseBash(program, child);
	    } else if (tag.equals("python")) {
		parsePython(program, child);
	    } else if (tag.equals("clojure")) {
		parseClojure(program, child);
	    } else if (tag.equals("perl")) {
		parsePerl(program, child);
	    } else {
		throw new ParseException("Unknown tag: " + tag);
	    }
	}
	return program;
    }

    private void parseBash(final Program program, Element element) throws ParseException {
	String id = assertAttribute(element, "id");
	BashScript script = new BashScript(id) {
		public Program getProgram() {
		    return program;
		}
	    };
	parseScript(script, element);
	program.withScript(id, script);
    }

    private void parsePython(final Program program, Element element) throws ParseException {
	String id = assertAttribute(element, "id");
	PythonScript script = new PythonScript(id) {
		public Program getProgram() {
		    return program;
		}
	    };
	parseScript(script, element);
	program.withScript(id, script);
    }

    private void parseClojure(final Program program, Element element) throws ParseException {
	String id = assertAttribute(element, "id");
	ClojureScript script = new ClojureScript(id) {
		public Program getProgram() {
		    return program;
		}
	    };
	parseScript(script, element);
	program.withScript(id, script);
    }

    private void parsePerl(final Program program, Element element) throws ParseException {
	String id = assertAttribute(element, "id");
	PerlScript script = new PerlScript(id) {
		public Program getProgram() {
		    return program;
		}
	    };
	parseScript(script, element);
	program.withScript(id, script);
    }

    private <S extends Script> void parseScript(S script, Element element) throws ParseException {
	for (Node child : new NodeIterable(element)) {
	    if (child.getNodeType() == Node.TEXT_NODE) {
		// append text piece
		script.withText(child.getNodeValue());
	    } else if (child.getNodeType() == Node.CDATA_SECTION_NODE) {
		// append text piece
		script.withText(child.getNodeValue());
	    } else if (child.getNodeType() == Node.ELEMENT_NODE) {
		Element childElement = (Element) child;
		if (childElement.getTagName().equals("run")) {
		    parseRun(childElement, script);
		} else if (childElement.getTagName().equals("param")) {
		    String paramId = parseParam(childElement);
		    script.withParam(paramId);
		} else {
		    throw new ParseException("Unexpected child element: " + childElement.getTagName());
		}
	    }
	}
    }

    private void parseRun(Element element, Script script) throws ParseException {
	String id = assertAttribute(element, "id");
	RunBuilder runBuilder = script.withRun(id);
	for (Element child : getElementChildren(element)) {
	    if (child.getTagName().equals("arg")) {
		parseArg(child, runBuilder);
	    } else {
		// unexpected
		throw new ParseException("Unexpected child element: " + child.getTagName());
	    }
	}
	runBuilder.build();
    }

    private void parseArg(Element element, RunBuilder runBuilder) throws ParseException {
	String id = assertAttribute(element, "id");
	// only allow text/cdata/param elements
	ArgBuilder argBuilder = runBuilder.withArg(id);
	for (Node child : new NodeIterable(element)) {
	    if (child.getNodeType() == Node.TEXT_NODE) {
		// text piece
		argBuilder.withText(child.getNodeValue());
	    } else if (child.getNodeType() == Node.CDATA_SECTION_NODE) {
		// text piece
		argBuilder.withText(child.getNodeValue());
	    } else if (child.getNodeType() == Node.ELEMENT_NODE) {
		Element childElement = (Element) child;
		if (childElement.getTagName().equals("param")) {
		    String paramId = parseParam(childElement);
		    argBuilder.withParam(paramId);
		} else {
		    // unexpected
		    throw new ParseException("Unexpected child element: " + childElement.getTagName());
		}
	    }
	}
	argBuilder.build();
    }

    private String parseParam(Element element) throws ParseException {
	if (!element.getTagName().equals("param")) {
	    throw new ParseException("Expecting param element, got " + element.getTagName());
	}
	return assertAttribute(element, "id");
    }

    private List<Element> getElementChildren(Element elt) {
	List<Element> elements = new ArrayList<>();
	for (Node child : new NodeIterable(elt)) {
	    if (child.getNodeType() == Node.ELEMENT_NODE) {
		elements.add((Element) child);
	    }
	}
	return elements;
    }

    private String assertAttribute(Element elt, String attr) throws ParseException {
	if (!elt.hasAttribute(attr)) {
	    throw new ParseException(String.format("Expecting attribute %s in element %s", attr, elt.getTagName()));
	}
	return elt.getAttribute(attr);
    }
}
