package scripty;

import scripty.Script.*;
import java.util.*;

public class ProgramRenderer {

    private final Program program;
    private String clojureJarPath;

    public ProgramRenderer(Program program) {
	this.program = program;
    }

    public void setClojureJarPath(String path) {
	this.clojureJarPath = path;
    }

    private void checkClojureJarPath() {
	for (String id : program.getScriptIds()) {
	    if (program.getScript(id) instanceof ClojureScript
		&& clojureJarPath == null) {
		throw new IllegalStateException("Must set clojure jar path");
	    }
	}
    }

    public String render() {
	final Set<String> ids = program.getScriptIds();
	final String main = program.getMain();
	checkClojureJarPath();

	if (!ids.contains(main)) {
	    throw new RuntimeException("Main script not found: " + main);
	}
	Script mainScript = program.getScript(main);

	StringBuilder builder = new StringBuilder("#!/bin/bash\n")
	    .append("__CLJJAR=" + clojureJarPath + "\n");

	for (String id : ids) {
	    Script script = program.getScript(id);
	    builder.append(String.format("\n__PROGRAM_%s=$(cat<<\"__EOF\"\n%s\n__EOF\n);\n", id, script.accept(renderVisitor)));
	}

	builder.append("\nenv __CLJJAR=\"$__CLJJAR\"");
	for (String id : ids) {
	    builder.append(String.format(" __PROGRAM_%s=\"$__PROGRAM_%s\"", id, id));
	}
	return builder.append(String.format(" %s ;", mainScript.accept(mainVisitor))).toString();
    }

    // program

    private static String renderArg(Arg arg, PieceVisitor<String> visitor) {
	StringBuilder builder = new StringBuilder();
	for (Piece piece : arg.getPieces()) {
	    builder.append(piece.accept(visitor));
	}
	return builder.toString();
    }

    private static final ScriptVisitor<String> mainVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    return String.format("bash -c \"$__PROGRAM_%s\"", script.getId());
	}
	public String visit(PythonScript script) {
	    return String.format("python -c \"$__PROGRAM_%s\"", script.getId());
	}
	public String visit(ClojureScript script) {
	    return String.format("java -jar \"$__CLJJAR\" -e \"$__PROGRAM_%s\"", script.getId());
	}
	public String visit(PerlScript script) {
	    return String.format("perl -e \"$__PROGRAM_%s\"", script.getId());
	}
    };

    private static final ScriptVisitor<String> renderVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    StringBuilder builder = new StringBuilder();
	    for (Piece piece : script.getPieces()) {
		builder.append(piece.accept(bashPieceVisitor));
	    }
	    return builder.toString();
	}
	public String visit(PythonScript script) {
	    StringBuilder builder = new StringBuilder("import os\nimport subprocess\n");
	    for (Piece piece : script.getPieces()) {
		builder.append(piece.accept(pythonPieceVisitor));
	    }
	    return builder.toString();
	}
	public String visit(ClojureScript script) {
	    StringBuilder builder = new StringBuilder();
	    for (Piece piece : script.getPieces()) {
		builder.append(piece.accept(clojurePieceVisitor));
	    }
	    return builder.toString();
	}
	public String visit(PerlScript script) {
	    StringBuilder builder = new StringBuilder();
	    for (Piece piece : script.getPieces()) {
		builder.append(piece.accept(perlPieceVisitor));
	    }
	    return builder.toString();
	}
    };

    // bash
    private static final PieceVisitor<String> bashPieceVisitor = new PieceVisitor<String>() {
	public String visit(Text text) {
	    return text.getText();
	}
	public String visit(Param param) {
	    return String.format("$(printenv __arg%s)", param.getIndex());
	}
	public String visit(Run run) {
	    Script runScript = run.getScript().getProgram().getScript(run);
	    StringBuilder argBuilder = new StringBuilder("env");
	    for (Arg arg : run.getArgs()) {
		Integer index = runScript.getParamIndex(arg.getId());
		argBuilder.append(String.format(" __arg%s=\"$(echo %s)\"", index, renderArg(arg, bashPieceVisitor)));
	    }
	    return String.format("$(%s %s)", argBuilder.toString(), runScript.accept(bashRunVisitor));
	}
    };

    private static final ScriptVisitor<String> bashRunVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    return String.format("bash -c \"$(printenv __PROGRAM_%s)\"", script.getId());
	}
	public String visit(PythonScript script) {
	    return String.format("python -c \"$(printenv __PROGRAM_%s)\"", script.getId());
	}
	public String visit(ClojureScript script) {
	    return String.format("java -jar \"$(printenv __CLJJAR)\" -e \"$(printenv __PROGRAM_%s)\"", script.getId());
	}
	public String visit(PerlScript script) {
	    return String.format("perl -e \"$(printenv __PROGRAM_%s)\"", script.getId());
	}
    };

    // python

    private static final PieceVisitor<String> pythonPieceVisitor = new PieceVisitor<String>() {
	public String visit(Text text) {
	    return text.getText();
	}
	public String visit(Param param) {
	    return String.format("os.environ['__arg%s']", param.getIndex());
	}
	public String visit(Run run) {
	    Script runScript = run.getScript().getProgram().getScript(run);
	    StringBuilder argBuilder = new StringBuilder("'env'");
	    for (Arg arg : run.getArgs()) {
		Integer index = runScript.getParamIndex(arg.getId());
		argBuilder.append(String.format(", '__arg%s=' + str(%s)", index, renderArg(arg, pythonPieceVisitor)));
	    }
	    return String.format("subprocess.Popen([%s, %s], stdout=subprocess.PIPE).communicate()[0]",
				 argBuilder.toString(),
				 runScript.accept(pythonRunVisitor));
	}
    };

    private static final ScriptVisitor<String> pythonRunVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    return String.format("'bash', '-c', os.environ['__PROGRAM_%s']", script.getId());
	}
	public String visit(PythonScript script) {
	    return String.format("'python', '-c', os.environ['__PROGRAM_%s']", script.getId());
	}
	public String visit(ClojureScript script) {
	    return String.format("'java', '-jar', os.environ['__CLJJAR'], '-e', os.environ['__PROGRAM_%s']", script.getId());
	}
	public String visit(PerlScript script) {
	    return String.format("'perl', '-e', os.environ['__PROGRAM_%s']", script.getId());
	}
    };

    // clojure

    private static final PieceVisitor<String> clojurePieceVisitor = new PieceVisitor<String>() {
	public String visit(Text text) {
	    return text.getText();
	}
	public String visit(Param param) {
	    return String.format("(java.lang.System/getenv \"__arg%s\")", param.getIndex());    
	}
	public String visit(Run run) {
	    Script runScript = run.getScript().getProgram().getScript(run);
	    StringBuilder argBuilder = new StringBuilder("\"env\"");
	    for (Arg arg : run.getArgs()) {
		Integer index = runScript.getParamIndex(arg.getId());
		argBuilder.append(String.format(" (str \"__arg%s=\" %s)", index, renderArg(arg, clojurePieceVisitor)));
	    }
	    return String.format("(slurp (.getInputStream (.exec (java.lang.Runtime/getRuntime) [%s %s] nil nil)))",
				 argBuilder.toString(),
				 runScript.accept(clojureRunVisitor));
	}
    };

    private static final ScriptVisitor<String> clojureRunVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    return String.format("[\"bash\", \"-c\", (java.lang.System/getenv \"__PROGRAM_%s\")]", script.getId());
	}
	public String visit(PythonScript script) {
	    return String.format("[\"python\", \"-c\", (java.lang.System/getenv \"__PROGRAM_%s\")]", script.getId());
	}
	public String visit(ClojureScript script) {
	    return String.format("[\"java\", \"-jar\", (java.lang.System/getenv \"__CLJJAR\"), \"-e\", (java.lang.System/getenv \"__PROGRAM_%s\")]", script.getId());
	}
	public String visit(PerlScript script) {
	    return String.format("[\"perl\", \"-e\", (java.lang.System/getenv \"__PROGRAM_%s\")]", script.getId());
	}
    };

    // perl

    private static final PieceVisitor<String> perlPieceVisitor = new PieceVisitor<String>() {
	public String visit(Text text) {
	    return text.getText();
	}
	public String visit(Param param) {
	    return String.format("$ENV{\"__arg%s\"}", param.getIndex());    
	}
	public String visit(Run run) {
	    Script runScript = run.getScript().getProgram().getScript(run);
	    StringBuilder argBuilder = new StringBuilder("\"env\"");
	    for (Arg arg : run.getArgs()) {
		Integer index = runScript.getParamIndex(arg.getId());
		argBuilder.append(String.format(", \"__arg%s=\" . (%s)", index, renderArg(arg, perlPieceVisitor)));
	    }
	    return String.format("system((%s, %s))", argBuilder.toString(), runScript.accept(perlRunVisitor));
	}
    };

    private static final ScriptVisitor<String> perlRunVisitor = new ScriptVisitor<String>() {
	public String visit(BashScript script) {
	    return String.format("\"bash\", \"-c\", $ENV{\"__PROGRAM_%s\"}", script.getId());
	}
	public String visit(PythonScript script) {
	    return String.format("\"python\", \"-c\", $ENV{\"__PROGRAM_%s\"}", script.getId());
	}
	public String visit(ClojureScript script) {
	    return String.format("\"java\", \"-jar\", $ENV{\"__CLJJAR\"}, \"-e\", $ENV{\"__PROGRAM_%s\"}", script.getId());
	}
	public String visit(PerlScript script) {
	    return String.format("\"perl\", \"-e\", $ENV{\"__PROGRAM_%s\"}", script.getId());
	}
    };
}
