package scripty;

import java.util.*;

public abstract class Script {

    public interface Piece {
	<T> T accept(PieceVisitor<T> visitor);
	Script getScript();
    }

    public interface PieceVisitor<T> {
	T visit(Text text);
	T visit(Param param);
	T visit(Run run);
    }

    public class Text implements Piece {
	private final String text;
	private Text(String text) {
	    this.text = text;
	}
	public String getText() {
	    return text;
	}
	public <T> T accept(PieceVisitor<T> visitor) {
	    return visitor.visit(this);
	}
	public Script getScript() {
	    return Script.this;
	}
    }

    public class Param implements Piece {
	private final String id;
	private final int index;
	private Param(String id, int index) {
	    this.id = id;
	    this.index = index;
	}
	public String getId() {
	    return id;
	}
	public int getIndex() {
	    return index;
	}
	public <T> T accept(PieceVisitor<T> visitor) {
	    return visitor.visit(this);
	}
	public Script getScript() {
	    return Script.this;
	}
    }

    public class Run implements Piece {
	private final String id;
	private final List<Arg> args;
	private Run(String id, List<Arg> args) {
	    this.id = id;
	    this.args = args;
	}
	public String getId() {
	    return this.id;
	}
	public List<Arg> getArgs() {
	    return args;
	}
	public <T> T accept(PieceVisitor<T> visitor) {
	    return visitor.visit(this);
	}
	public Script getScript() {
	    return Script.this;
	}
    }

    public class Arg {
	private final String id;
	private final List<Piece> pieces;
	private Arg(String id, List<Piece> pieces) {
	    this.id = id;
	    this.pieces = pieces;
	}
	public String getId() {
	    return id;
	}
	public List<Piece> getPieces() {
	    return pieces;
	}
    }

    protected final String id;
    protected final List<Piece> pieces;
    protected final List<Param> numberedParams;
    protected int nextParamIndex = 0;

    protected Script(String id) {
	this.id = id;
	this.pieces = new ArrayList<>();
	this.numberedParams = new ArrayList<>();
    }

    protected abstract Program getProgram();
    protected abstract <T> T accept(ScriptVisitor<T> visitor);

    public String getId() {
	return id;
    }

    public List<Piece> getPieces() {
	return pieces;
    }

    public Integer getParamIndex(String id) {
	for (Param param : numberedParams) {
	    if (param.getId().equals(id)) {
		return param.getIndex();
	    }
	}
	return null;
    }

    public Script withText(String text) {
	this.pieces.add(new Text(text));
	return this;
    }

    public Script withParam(String id) {
	Param param = new Param(id, nextParamIndex);
	this.numberedParams.add(param);
	this.pieces.add(param);
	nextParamIndex++;
	return this;
    }

    public RunBuilder withRun(String id) {
	return new RunBuilder(id);
    }

    protected void assertValidRun(Run run) {
	String callerId = run.getScript().getId();
	if (!run.getId().equals(getId())) {
	    throw new RuntimeException("Run is for wrong script id: " + getId() + " != " + run.getId());
	}
	Set<String> runNames = new HashSet<>();
	for (Arg arg : run.getArgs()) {
	    if (getParamIndex(arg.getId()) == null) {
		throw new RuntimeException("Script " + callerId + " calls script " + getId() + " with unknown arg: " + arg.getId());
	    }
	    runNames.add(arg.getId());
	}
	for (Param param : numberedParams) {
	    if (!runNames.contains(param.getId())) {
		throw new RuntimeException("Script " + callerId + " call script " + getId() + " without param: " + param.getId());
	    }
	}
    }

    protected int getRunParamIndex(Script runScript, String argId) {
	Integer index = runScript.getParamIndex(argId);
	if (index == null) {
	    throw new RuntimeException("Script " + getId() + " attempting to call script " + runScript.getId() + " with unknown param id: " + argId);
	} else {
	    return index;
	}
    }

    public class RunBuilder {
	private final String id;
	private final List<Arg> args = new ArrayList<>();

	private RunBuilder(String id) {
	    this.id = id;
	}

	public ArgBuilder withArg(String id) {
	    return new ArgBuilder(id);
	}

	public Script build() {
	    Script.this.pieces.add(new Run(id, args));
	    return Script.this;
	}

	public class ArgBuilder {
	    private final List<Piece> pieces = new ArrayList<>();
	    private final String id;
	
	    private ArgBuilder(String id) {
		this.id = id;
	    }

	    public ArgBuilder withText(String text) {
		this.pieces.add(new Text(text));
		return this;
	    }

	    public ArgBuilder withParam(String id) {
		Param param = new Param(id, nextParamIndex);
		numberedParams.add(param);
		this.pieces.add(param);
		nextParamIndex++;
		return this;
	    }

	    public RunBuilder build() {
		if (pieces.isEmpty()) {
		    throw new RuntimeException("No pieces in arg");
		}
		RunBuilder.this.args.add(new Arg(id, pieces));
		return RunBuilder.this;
	    }
	}
    }
}
